/*
 * physical constants used by Cloudy, taken from
 * Mohr P.J. & Taylor B.N., 1998 Codata, see Reviews of Modern Physics, Vol. 72, No. 2, 2000
 * or http://www.physics.nist.gov/constants
 */
/*#include "physconst.h"*/

/*********************************************************************
 * first come math constants                                         *
 *********************************************************************/

/* pi */
#define	PI	3.141592653589793238462643

/* 4*pi */
#define	PI4	12.56637061435917295385057

/* sqrt(pi) */
#define	SQRTPI	1.772453850905516027298167

/*********************************************************************
 * astronomical constants go here                                    *
 *********************************************************************/

/* solar mass in grams */
#define	SOLAR_MASS	1.9892e33

/* parsec in cm */
#define	PARSEC	        3.085678e18

/*********************************************************************
 * fundamental constants go next, eventually rest should be defined  *
 * in terms of these                                                 *
 *********************************************************************/

/* atomic mass unit, grams */
#define	ATOMIC_MASS_UNIT    1.66053873e-24

/* electron mass, grams */
#define ELECTRON_MASS	    9.10938188e-28

/* proton mass, grams */
#define PROTON_MASS	    1.67262158e-24

/* this is the Boltzmann factor, erg/K */
#define	BOLTZMANN	    1.3806503e-16

/* speed of light, cm/s */
#define	SPEEDLIGHT	    2.99792458e10

/* Planck's constant */
#define	HPLANCK	            6.62606876e-27

/* elementary charge, in C */
#define ELEM_CHARGE         1.602176462e-19

/* infinite mass rydberg constant, in cm^-1 */
#define RYD_INF             1.0973731568549e5

/* ionization potential of real hydrogen atom, in inf mass ryd, based on Codata 98,
 * uncertainty 24e-12,  calculated by Peter van Hoof */
#define	HIONPOT	            0.999466508219

/*********************************************************************
 * below here should be derived constants                            *
 *                                                                   *
 * NB - explicit values in comments are approximate                  *
 *      and are not maintained !                                     *
 *********************************************************************/

/* electric constant, in F/m, 8.854e-12 */
#define ELECTRIC_CONST      (1.e11/PI4/SPEEDLIGHT/SPEEDLIGHT)

/* this is the factor that appears in front of boltzmann factor to get
 * LTE level populations for hydrogenic ions.  It is givin in the
 * first parts of section 5 of part 2 of hazy, and is actually
 * ( planck^2 / (2 pi m_e k ) )^3/2, but cannot evaluate powers here,
 * must raise this to 3/2 when used, HION_LTE_POP = 5.556e-11  */
#define HION_LTE_POP	    (HPLANCK*HPLANCK/2./PI/BOLTZMANN/ELECTRON_MASS)

/* SAHA2 is ( h^2/2/pi/m/k )^3, is correct constant for free electron
 * NB - the exponent should actually be 3/2, so take sqrt(SAHA2) before use !!
 * SAHA = 4.141e-16, SAHA2 = 1.7151e-31 */
#define SAHA2               (HION_LTE_POP*HION_LTE_POP*HION_LTE_POP)

/* 1 cm wavenumber in ergs, 1.9864e-16 */
#define	ERG1CM	            (HPLANCK*SPEEDLIGHT)

/* degrees kelvin per unit wavenumber, 1.4388 */
#define	T1CM	            (HPLANCK*SPEEDLIGHT/BOLTZMANN)

/* conversion factor from wavenumbers to degrees kelvin - never used, same as T1CM */
/*#define	WAVNKELV	    (HPLANCK*SPEEDLIGHT/BOLTZMANN)*/

/* number of Ryd per wavenumber, 9.11267e-6 */
#define	WAVNRYD	            (1./RYD_INF)

/* 1 infinite mass Rydberg unit in Angstrom, 911.2671 */
#define	RYDLAM	            (1.e8/RYD_INF)

/* inf mass energy of 1 Ryd in ergs, 2.180e-11 */
#define	EN1RYD	            (HPLANCK*SPEEDLIGHT*RYD_INF)

/* the temperature of 1 Rydberg */
/* te1ryd is h/k is temp of 1 Rydberg, 1.579e5 */
#define	TE1RYD	            (HPLANCK*SPEEDLIGHT*RYD_INF/BOLTZMANN)

/* 1 eV in Kelvin, 1.1604e4 */
#define	EVDEGK	            (ELEM_CHARGE*1.e7/BOLTZMANN)

/* 1 Ryd (inf mass nucleus) in eV, 13.606 */
#define	EVRYD	            (HPLANCK*SPEEDLIGHT*RYD_INF/ELEM_CHARGE*1.e-7)

/* frequency of one Ryd for infinite mass nuclei, 3.289842e15 */
#define	FR1RYD	            (SPEEDLIGHT*RYD_INF)

/*2 h FR1RYD^3 / c^2 for infinite mass nucleus, 0.5250 */
#define	HNU3C2	            (2.*HPLANCK*SPEEDLIGHT*RYD_INF*RYD_INF*RYD_INF)

/* frequency of ionization potential of Hydrogen, 3.288087e15 - never used */
#define	FR1RYDHYD           (SPEEDLIGHT*RYD_INF*HIONPOT)

/* wavelength (A) of ionization potential of Hydrogen, 911.7535 - never used */
#define	RYDLAMHYD           (RYDLAM/HIONPOT)
